/**
 *  \file RMF/alias_decorators.h
 *  \brief Helper functions for manipulating RMF files.
 *
 *  Copyright 2007-2013 IMP Inventors. All rights reserved.
 *
 */

#ifndef RMF_ALIAS_DECORATORS_H
#define RMF_ALIAS_DECORATORS_H

#include <RMF/config.h>
#include <RMF/infrastructure_macros.h>
#include <RMF/NodeHandle.h>
#include <RMF/FileHandle.h>
#include <RMF/Decorator.h>
#include <RMF/Factory.h>
#include <RMF/constants.h>
#include <RMF/internal/utility.h>
#include <RMF/internal/paths.h>
#include <boost/array.hpp>
#include <boost/lexical_cast.hpp>

RMF_ENABLE_WARNINGS
namespace RMF {

/** Store a reference to another node as an alias.

       See also Alias and AliasConstFactory.
     */
    class AliasConst:
        public Decorator<NodeConstHandle> {
    friend class AliasConstFactory;
    std::string get_name() const {return "Alias";}
    private:
    typedef Decorator<NodeConstHandle> P;
    NodeIDKey aliased_;
    AliasConst(NodeConstHandle nh,
                  NodeIDKey aliased):
       P(nh),
aliased_(aliased) {
    ;
    }
    public:
    /** The node that is referenced. */
NodeConstHandle get_aliased() const {
  try {
  NodeID id;
   id= get_node().get_value(aliased_);
  return get_node().get_file().get_node_from_id(id);
  } RMF_DECORATOR_CATCH( );
}
    static std::string get_decorator_type_name() {
         return "AliasConst";
    }
    };

    typedef std::vector<AliasConst>
            AliasConsts;

/** Store a reference to another node as an alias.

       See also AliasConst and AliasFactory.
     */
    class Alias:
        public Decorator<NodeHandle> {
    friend class AliasFactory;
    std::string get_name() const {return "Alias";}
    private:
    typedef Decorator<NodeHandle> P;
    NodeIDKey aliased_;
    Alias(NodeHandle nh,
                  NodeIDKey aliased):
       P(nh),
aliased_(aliased) {
    ;
    }
    public:
    /** The node that is referenced. */
NodeHandle get_aliased() const {
  try {
  NodeID id;
   id= get_node().get_value(aliased_);
  return get_node().get_file().get_node_from_id(id);
  } RMF_DECORATOR_CATCH( );
}
/** The node that is referenced. */
void set_aliased(NodeConstHandle v) {
  try {
    get_node().set_value(aliased_, v.get_id());
  } RMF_DECORATOR_CATCH( );
}
    static std::string get_decorator_type_name() {
         return "Alias";
    }
    };

    typedef std::vector<Alias>
            Aliass;

/** Create decorators of type Alias.

       See also Alias and AliasConstFactory.
    */
    class AliasFactory:
       public Factory<FileHandle>
                 {
    private:
    typedef Factory<FileHandle> P;
    NodeIDKey aliased_;
    public:
    AliasFactory(FileHandle fh):
      P(){
    Category cat = fh.get_category("alias");
RMF_UNUSED(cat);
aliased_=fh.get_key<NodeIDTraits>(cat,
        "aliased");;
    }
    /** Get a Alias for nh.*/
    Alias get(NodeHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::ALIAS), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Alias");;
      return Alias(nh, aliased_);
    }
    /** Check whether nh has all the attributes required to be a
        Alias.*/
    bool get_is(NodeHandle nh) const {
      return (nh.get_type() == RMF::ALIAS)
    && nh.get_has_value(aliased_);
    }
    };

    typedef std::vector<AliasFactory>
            AliasFactories;

/** Create decorators of type Alias.

       See also AliasConst and AliasFactory.
    */
    class AliasConstFactory:
       public Factory<FileConstHandle>
                 {
    private:
    typedef Factory<FileConstHandle> P;
    NodeIDKey aliased_;
    public:
    AliasConstFactory(FileConstHandle fh):
      P(){
    Category cat = fh.get_category("alias");
RMF_UNUSED(cat);
aliased_=fh.get_key<NodeIDTraits>(cat,
        "aliased");;
    }
    /** Get a AliasConst for nh.*/
    AliasConst get(NodeConstHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::ALIAS), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Alias");;
      return AliasConst(nh, aliased_);
    }
    /** Check whether nh has all the attributes required to be a
        AliasConst.*/
    bool get_is(NodeConstHandle nh) const {
      return (nh.get_type() == RMF::ALIAS)
    && nh.get_has_value(aliased_);
    }
    };

    typedef std::vector<AliasConstFactory>
            AliasConstFactories;

} /* namespace RMF */
RMF_DISABLE_WARNINGS

#endif /* RMF_ALIAS_DECORATORS_H */
