/**
 *  \file RMF/feature_decorators.h
 *  \brief Helper functions for manipulating RMF files.
 *
 *  Copyright 2007-2013 IMP Inventors. All rights reserved.
 *
 */

#ifndef RMF_FEATURE_DECORATORS_H
#define RMF_FEATURE_DECORATORS_H

#include <RMF/config.h>
#include <RMF/infrastructure_macros.h>
#include <RMF/NodeHandle.h>
#include <RMF/FileHandle.h>
#include <RMF/Decorator.h>
#include <RMF/Factory.h>
#include <RMF/constants.h>
#include <RMF/internal/utility.h>
#include <RMF/internal/paths.h>
#include <boost/array.hpp>
#include <boost/lexical_cast.hpp>

RMF_ENABLE_WARNINGS
namespace RMF {

/** Associate a score with some set of particles. If the score is an energy, it should be in kJ/mol.

       See also Score and ScoreConstFactory.
     */
    class ScoreConst:
        public Decorator<NodeConstHandle> {
    friend class ScoreConstFactory;
    std::string get_name() const {return "Score";}
    private:
    typedef Decorator<NodeConstHandle> P;
    AliasConstFactory representation_;
FloatKey score_;
    ScoreConst(NodeConstHandle nh,
                  AliasConstFactory representation,
FloatKey score):
       P(nh),
representation_(representation),
score_(score) {
    ;
    }
    public:
    /** The various components of this score node. */
NodeConstHandles get_representation() const {
  try{
  NodeConstHandles typed=get_node().get_children();
  NodeConstHandles ret;
  for (unsigned int i=0; i< typed.size(); ++i) {
     if (representation_.get_is(typed[i])) {
        ret.push_back(representation_.get(typed[i]).get_aliased());
     }
  }
  return ret;
  } RMF_DECORATOR_CATCH( );
}
/** The score. */
Float get_score() const {
  try{
  return P::get_value(score_);
  } RMF_DECORATOR_CATCH( );
}
/** The score. */
Floats get_all_scores() const {
  try {
  return P::get_all_values(score_);
  } RMF_DECORATOR_CATCH( );
}
    static std::string get_decorator_type_name() {
         return "ScoreConst";
    }
    };

    typedef std::vector<ScoreConst>
            ScoreConsts;

/** Associate a score with some set of particles. If the score is an energy, it should be in kJ/mol.

       See also ScoreConst and ScoreFactory.
     */
    class Score:
        public Decorator<NodeHandle> {
    friend class ScoreFactory;
    std::string get_name() const {return "Score";}
    private:
    typedef Decorator<NodeHandle> P;
    AliasFactory representation_;
FloatKey score_;
    Score(NodeHandle nh,
                  AliasFactory representation,
FloatKey score):
       P(nh),
representation_(representation),
score_(score) {
    ;
    }
    public:
    /** The various components of this score node. */
NodeHandles get_representation() const {
  try{
  NodeHandles typed=get_node().get_children();
  NodeHandles ret;
  for (unsigned int i=0; i< typed.size(); ++i) {
     if (representation_.get_is(typed[i])) {
        ret.push_back(representation_.get(typed[i]).get_aliased());
     }
  }
  return ret;
  } RMF_DECORATOR_CATCH( );
}
/** The various components of this score node. */
void set_representation(NodeConstHandles v) {
  try{
   for (unsigned int i=0; i< v.size(); ++i) {
       internal::add_child_alias(representation_, get_node(), v[i]);
   }
  } RMF_DECORATOR_CATCH( );
}
/** The various components of this score node. */
void set_representation(NodeHandles v) {
  try{
   for (unsigned int i=0; i< v.size(); ++i) {
       internal::add_child_alias(representation_, get_node(), v[i]);
   }
  } RMF_DECORATOR_CATCH( );
}
/** The score. */
Float get_score() const {
  try{
  return P::get_value(score_);
  } RMF_DECORATOR_CATCH( );
}
/** The score. */
Floats get_all_scores() const {
  try {
  return P::get_all_values(score_);
  } RMF_DECORATOR_CATCH( );
}
/** The score. */
void set_score(Float v) {
  try {
  P::set_value(score_,
            v);
  } RMF_DECORATOR_CATCH( );
}
    static std::string get_decorator_type_name() {
         return "Score";
    }
    };

    typedef std::vector<Score>
            Scores;

/** Create decorators of type Score.

       See also Score and ScoreConstFactory.
    */
    class ScoreFactory:
       public Factory<FileHandle>
                 {
    private:
    typedef Factory<FileHandle> P;
    AliasFactory representation_;
FloatKey score_;
    public:
    ScoreFactory(FileHandle fh):
      P(), representation_(fh){
    Category cat = fh.get_category("feature");
RMF_UNUSED(cat);
score_=fh.get_key<FloatTraits>(cat,
        "score");;
    }
    /** Get a Score for nh.*/
    Score get(NodeHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::FEATURE), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Score");;
      return Score(nh, representation_,
score_);
    }
    /** Check whether nh has all the attributes required to be a
        Score.*/
    bool get_is(NodeHandle nh) const {
      return (nh.get_type() == RMF::FEATURE)
    && nh.get_has_value(score_);
    }
    };

    typedef std::vector<ScoreFactory>
            ScoreFactories;

/** Create decorators of type Score.

       See also ScoreConst and ScoreFactory.
    */
    class ScoreConstFactory:
       public Factory<FileConstHandle>
                 {
    private:
    typedef Factory<FileConstHandle> P;
    AliasConstFactory representation_;
FloatKey score_;
    public:
    ScoreConstFactory(FileConstHandle fh):
      P(), representation_(fh){
    Category cat = fh.get_category("feature");
RMF_UNUSED(cat);
score_=fh.get_key<FloatTraits>(cat,
        "score");;
    }
    /** Get a ScoreConst for nh.*/
    ScoreConst get(NodeConstHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::FEATURE), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Score");;
      return ScoreConst(nh, representation_,
score_);
    }
    /** Check whether nh has all the attributes required to be a
        ScoreConst.*/
    bool get_is(NodeConstHandle nh) const {
      return (nh.get_type() == RMF::FEATURE)
    && nh.get_has_value(score_);
    }
    };

    typedef std::vector<ScoreConstFactory>
            ScoreConstFactories;

} /* namespace RMF */
RMF_DISABLE_WARNINGS

#endif /* RMF_FEATURE_DECORATORS_H */
