/**
 *  \file RMF/physics_decorators.h
 *  \brief Helper functions for manipulating RMF files.
 *
 *  Copyright 2007-2013 IMP Inventors. All rights reserved.
 *
 */

#ifndef RMF_PHYSICS_DECORATORS_H
#define RMF_PHYSICS_DECORATORS_H

#include <RMF/config.h>
#include <RMF/infrastructure_macros.h>
#include <RMF/NodeHandle.h>
#include <RMF/FileHandle.h>
#include <RMF/Decorator.h>
#include <RMF/Factory.h>
#include <RMF/constants.h>
#include <RMF/internal/utility.h>
#include <RMF/internal/paths.h>
#include <boost/array.hpp>
#include <boost/lexical_cast.hpp>
#include "alias_decorators.h" 

RMF_ENABLE_WARNINGS
namespace RMF {

/** These particles has associated coordinates and radius information.

       See also Particle and ParticleConstFactory.
     */
    class ParticleConst:
        public Decorator<NodeConstHandle> {
    friend class ParticleConstFactory;
    std::string get_name() const {return "Particle";}
    private:
    typedef Decorator<NodeConstHandle> P;
    FloatKeys coordinates_;
FloatKey radius_;
FloatKey mass_;
    ParticleConst(NodeConstHandle nh,
                  FloatKeys coordinates,
FloatKey radius,
FloatKey mass):
       P(nh),
coordinates_(coordinates),
radius_(radius),
mass_(mass) {
    ;
    }
    public:
    /** The coordinates in angstroms. */
Floats get_coordinates() const {
             return P::get_values(coordinates_);
           }
/** The radius in angstroms. */
Float get_radius() const {
  try{
  return P::get_value(radius_);
  } RMF_DECORATOR_CATCH( );
}
/** The radius in angstroms. */
Floats get_all_radiuss() const {
  try {
  return P::get_all_values(radius_);
  } RMF_DECORATOR_CATCH( );
}
/** The mass in Daltons. */
Float get_mass() const {
  try{
  return P::get_value(mass_);
  } RMF_DECORATOR_CATCH( );
}
/** The mass in Daltons. */
Floats get_all_masss() const {
  try {
  return P::get_all_values(mass_);
  } RMF_DECORATOR_CATCH( );
}
    static std::string get_decorator_type_name() {
         return "ParticleConst";
    }
    };

    typedef std::vector<ParticleConst>
            ParticleConsts;

/** These particles has associated coordinates and radius information.

       See also ParticleConst and ParticleFactory.
     */
    class Particle:
        public Decorator<NodeHandle> {
    friend class ParticleFactory;
    std::string get_name() const {return "Particle";}
    private:
    typedef Decorator<NodeHandle> P;
    FloatKeys coordinates_;
FloatKey radius_;
FloatKey mass_;
    Particle(NodeHandle nh,
                  FloatKeys coordinates,
FloatKey radius,
FloatKey mass):
       P(nh),
coordinates_(coordinates),
radius_(radius),
mass_(mass) {
    ;
    }
    public:
    /** The coordinates in angstroms. */
Floats get_coordinates() const {
             return P::get_values(coordinates_);
           }
/** The coordinates in angstroms. */
    void set_coordinates(const Floats &v) {
           P::set_values(coordinates_, v);
        }
/** The radius in angstroms. */
Float get_radius() const {
  try{
  return P::get_value(radius_);
  } RMF_DECORATOR_CATCH( );
}
/** The radius in angstroms. */
Floats get_all_radiuss() const {
  try {
  return P::get_all_values(radius_);
  } RMF_DECORATOR_CATCH( );
}
/** The radius in angstroms. */
void set_radius(Float v) {
  try {
  P::set_value(radius_,
            v);
  } RMF_DECORATOR_CATCH( );
}
/** The mass in Daltons. */
Float get_mass() const {
  try{
  return P::get_value(mass_);
  } RMF_DECORATOR_CATCH( );
}
/** The mass in Daltons. */
Floats get_all_masss() const {
  try {
  return P::get_all_values(mass_);
  } RMF_DECORATOR_CATCH( );
}
/** The mass in Daltons. */
void set_mass(Float v) {
  try {
  P::set_value(mass_,
            v);
  } RMF_DECORATOR_CATCH( );
}
    static std::string get_decorator_type_name() {
         return "Particle";
    }
    };

    typedef std::vector<Particle>
            Particles;

/** Create decorators of type Particle.

       See also Particle and ParticleConstFactory.
    */
    class ParticleFactory:
       public Factory<FileHandle>
                 {
    private:
    typedef Factory<FileHandle> P;
    FloatKeys coordinates_;
FloatKey radius_;
FloatKey mass_;
    public:
    ParticleFactory(FileHandle fh):
      P(){
    Category cat = fh.get_category("physics");
RMF_UNUSED(cat);
        Strings  coordinates_names;
        coordinates_names.push_back("cartesian x");
        coordinates_names.push_back("cartesian y");
        coordinates_names.push_back("cartesian z");
      coordinates_=fh.get_keys<FloatTraits>(cat,
        coordinates_names);
radius_=fh.get_key<FloatTraits>(cat,
        "radius");
mass_=fh.get_key<FloatTraits>(cat,
        "mass");;
    }
    /** Get a Particle for nh.*/
    Particle get(NodeHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Particle");;
      return Particle(nh, coordinates_,
radius_,
mass_);
    }
    /** Check whether nh has all the attributes required to be a
        Particle.*/
    bool get_is(NodeHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(coordinates_[0])
    && nh.get_has_value(radius_)
    && nh.get_has_value(mass_);
    }
    };

    typedef std::vector<ParticleFactory>
            ParticleFactories;

/** Create decorators of type Particle.

       See also ParticleConst and ParticleFactory.
    */
    class ParticleConstFactory:
       public Factory<FileConstHandle>
                 {
    private:
    typedef Factory<FileConstHandle> P;
    FloatKeys coordinates_;
FloatKey radius_;
FloatKey mass_;
    public:
    ParticleConstFactory(FileConstHandle fh):
      P(){
    Category cat = fh.get_category("physics");
RMF_UNUSED(cat);
        Strings  coordinates_names;
        coordinates_names.push_back("cartesian x");
        coordinates_names.push_back("cartesian y");
        coordinates_names.push_back("cartesian z");
      coordinates_=fh.get_keys<FloatTraits>(cat,
        coordinates_names);
radius_=fh.get_key<FloatTraits>(cat,
        "radius");
mass_=fh.get_key<FloatTraits>(cat,
        "mass");;
    }
    /** Get a ParticleConst for nh.*/
    ParticleConst get(NodeConstHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Particle");;
      return ParticleConst(nh, coordinates_,
radius_,
mass_);
    }
    /** Check whether nh has all the attributes required to be a
        ParticleConst.*/
    bool get_is(NodeConstHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(coordinates_[0])
    && nh.get_has_value(radius_)
    && nh.get_has_value(mass_);
    }
    };

    typedef std::vector<ParticleConstFactory>
            ParticleConstFactories;

/** These particles have associated coordinates and radius information.

       See also IntermediateParticle and IntermediateParticleConstFactory.
     */
    class IntermediateParticleConst:
        public Decorator<NodeConstHandle> {
    friend class IntermediateParticleConstFactory;
    std::string get_name() const {return "IntermediateParticle";}
    private:
    typedef Decorator<NodeConstHandle> P;
    FloatKeys coordinates_;
FloatKey radius_;
    IntermediateParticleConst(NodeConstHandle nh,
                  FloatKeys coordinates,
FloatKey radius):
       P(nh),
coordinates_(coordinates),
radius_(radius) {
    ;
    }
    public:
    /** The coordinates in angstroms. */
Floats get_coordinates() const {
             return P::get_values(coordinates_);
           }
/** The radius in angstroms. */
Float get_radius() const {
  try{
  return P::get_value(radius_);
  } RMF_DECORATOR_CATCH( );
}
/** The radius in angstroms. */
Floats get_all_radiuss() const {
  try {
  return P::get_all_values(radius_);
  } RMF_DECORATOR_CATCH( );
}
    static std::string get_decorator_type_name() {
         return "IntermediateParticleConst";
    }
    };

    typedef std::vector<IntermediateParticleConst>
            IntermediateParticleConsts;

/** These particles have associated coordinates and radius information.

       See also IntermediateParticleConst and IntermediateParticleFactory.
     */
    class IntermediateParticle:
        public Decorator<NodeHandle> {
    friend class IntermediateParticleFactory;
    std::string get_name() const {return "IntermediateParticle";}
    private:
    typedef Decorator<NodeHandle> P;
    FloatKeys coordinates_;
FloatKey radius_;
    IntermediateParticle(NodeHandle nh,
                  FloatKeys coordinates,
FloatKey radius):
       P(nh),
coordinates_(coordinates),
radius_(radius) {
    ;
    }
    public:
    /** The coordinates in angstroms. */
Floats get_coordinates() const {
             return P::get_values(coordinates_);
           }
/** The coordinates in angstroms. */
    void set_coordinates(const Floats &v) {
           P::set_values(coordinates_, v);
        }
/** The radius in angstroms. */
Float get_radius() const {
  try{
  return P::get_value(radius_);
  } RMF_DECORATOR_CATCH( );
}
/** The radius in angstroms. */
Floats get_all_radiuss() const {
  try {
  return P::get_all_values(radius_);
  } RMF_DECORATOR_CATCH( );
}
/** The radius in angstroms. */
void set_radius(Float v) {
  try {
  P::set_value(radius_,
            v);
  } RMF_DECORATOR_CATCH( );
}
    static std::string get_decorator_type_name() {
         return "IntermediateParticle";
    }
    };

    typedef std::vector<IntermediateParticle>
            IntermediateParticles;

/** Create decorators of type IntermediateParticle.

       See also IntermediateParticle and IntermediateParticleConstFactory.
    */
    class IntermediateParticleFactory:
       public Factory<FileHandle>
                 {
    private:
    typedef Factory<FileHandle> P;
    FloatKeys coordinates_;
FloatKey radius_;
    public:
    IntermediateParticleFactory(FileHandle fh):
      P(){
    Category cat = fh.get_category("physics");
RMF_UNUSED(cat);
        Strings  coordinates_names;
        coordinates_names.push_back("cartesian x");
        coordinates_names.push_back("cartesian y");
        coordinates_names.push_back("cartesian z");
      coordinates_=fh.get_keys<FloatTraits>(cat,
        coordinates_names);
radius_=fh.get_key<FloatTraits>(cat,
        "radius");;
    }
    /** Get a IntermediateParticle for nh.*/
    IntermediateParticle get(NodeHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  IntermediateParticle");;
      return IntermediateParticle(nh, coordinates_,
radius_);
    }
    /** Check whether nh has all the attributes required to be a
        IntermediateParticle.*/
    bool get_is(NodeHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(coordinates_[0])
    && nh.get_has_value(radius_);
    }
    };

    typedef std::vector<IntermediateParticleFactory>
            IntermediateParticleFactories;

/** Create decorators of type IntermediateParticle.

       See also IntermediateParticleConst and IntermediateParticleFactory.
    */
    class IntermediateParticleConstFactory:
       public Factory<FileConstHandle>
                 {
    private:
    typedef Factory<FileConstHandle> P;
    FloatKeys coordinates_;
FloatKey radius_;
    public:
    IntermediateParticleConstFactory(FileConstHandle fh):
      P(){
    Category cat = fh.get_category("physics");
RMF_UNUSED(cat);
        Strings  coordinates_names;
        coordinates_names.push_back("cartesian x");
        coordinates_names.push_back("cartesian y");
        coordinates_names.push_back("cartesian z");
      coordinates_=fh.get_keys<FloatTraits>(cat,
        coordinates_names);
radius_=fh.get_key<FloatTraits>(cat,
        "radius");;
    }
    /** Get a IntermediateParticleConst for nh.*/
    IntermediateParticleConst get(NodeConstHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  IntermediateParticle");;
      return IntermediateParticleConst(nh, coordinates_,
radius_);
    }
    /** Check whether nh has all the attributes required to be a
        IntermediateParticleConst.*/
    bool get_is(NodeConstHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(coordinates_[0])
    && nh.get_has_value(radius_);
    }
    };

    typedef std::vector<IntermediateParticleConstFactory>
            IntermediateParticleConstFactories;

/** These particles have associated coordinates and orientation information.

       See also RigidParticle and RigidParticleConstFactory.
     */
    class RigidParticleConst:
        public Decorator<NodeConstHandle> {
    friend class RigidParticleConstFactory;
    std::string get_name() const {return "RigidParticle";}
    private:
    typedef Decorator<NodeConstHandle> P;
    FloatKeys orientation_;
FloatKeys coordinates_;
    RigidParticleConst(NodeConstHandle nh,
                  FloatKeys orientation,
FloatKeys coordinates):
       P(nh),
orientation_(orientation),
coordinates_(coordinates) {
    ;
    }
    public:
    /** The orientation as a quaternion. */
Floats get_orientation() const {
             return P::get_values(orientation_);
           }
/** The coordinates of the center in angstroms. */
Floats get_coordinates() const {
             return P::get_values(coordinates_);
           }
    static std::string get_decorator_type_name() {
         return "RigidParticleConst";
    }
    };

    typedef std::vector<RigidParticleConst>
            RigidParticleConsts;

/** These particles have associated coordinates and orientation information.

       See also RigidParticleConst and RigidParticleFactory.
     */
    class RigidParticle:
        public Decorator<NodeHandle> {
    friend class RigidParticleFactory;
    std::string get_name() const {return "RigidParticle";}
    private:
    typedef Decorator<NodeHandle> P;
    FloatKeys orientation_;
FloatKeys coordinates_;
    RigidParticle(NodeHandle nh,
                  FloatKeys orientation,
FloatKeys coordinates):
       P(nh),
orientation_(orientation),
coordinates_(coordinates) {
    ;
    }
    public:
    /** The orientation as a quaternion. */
Floats get_orientation() const {
             return P::get_values(orientation_);
           }
/** The orientation as a quaternion. */
    void set_orientation(const Floats &v) {
           P::set_values(orientation_, v);
        }
/** The coordinates of the center in angstroms. */
Floats get_coordinates() const {
             return P::get_values(coordinates_);
           }
/** The coordinates of the center in angstroms. */
    void set_coordinates(const Floats &v) {
           P::set_values(coordinates_, v);
        }
    static std::string get_decorator_type_name() {
         return "RigidParticle";
    }
    };

    typedef std::vector<RigidParticle>
            RigidParticles;

/** Create decorators of type RigidParticle.

       See also RigidParticle and RigidParticleConstFactory.
    */
    class RigidParticleFactory:
       public Factory<FileHandle>
                 {
    private:
    typedef Factory<FileHandle> P;
    FloatKeys orientation_;
FloatKeys coordinates_;
    public:
    RigidParticleFactory(FileHandle fh):
      P(){
    Category cat = fh.get_category("physics");
RMF_UNUSED(cat);
        Strings  orientation_names;
        orientation_names.push_back("orientation r");
        orientation_names.push_back("orientation i");
        orientation_names.push_back("orientation j");
        orientation_names.push_back("orientation k");
      orientation_=fh.get_keys<FloatTraits>(cat,
        orientation_names);
        Strings  coordinates_names;
        coordinates_names.push_back("cartesian x");
        coordinates_names.push_back("cartesian y");
        coordinates_names.push_back("cartesian z");
      coordinates_=fh.get_keys<FloatTraits>(cat,
        coordinates_names);;
    }
    /** Get a RigidParticle for nh.*/
    RigidParticle get(NodeHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  RigidParticle");;
      return RigidParticle(nh, orientation_,
coordinates_);
    }
    /** Check whether nh has all the attributes required to be a
        RigidParticle.*/
    bool get_is(NodeHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(orientation_[0])
    && nh.get_has_value(coordinates_[0]);
    }
    };

    typedef std::vector<RigidParticleFactory>
            RigidParticleFactories;

/** Create decorators of type RigidParticle.

       See also RigidParticleConst and RigidParticleFactory.
    */
    class RigidParticleConstFactory:
       public Factory<FileConstHandle>
                 {
    private:
    typedef Factory<FileConstHandle> P;
    FloatKeys orientation_;
FloatKeys coordinates_;
    public:
    RigidParticleConstFactory(FileConstHandle fh):
      P(){
    Category cat = fh.get_category("physics");
RMF_UNUSED(cat);
        Strings  orientation_names;
        orientation_names.push_back("orientation r");
        orientation_names.push_back("orientation i");
        orientation_names.push_back("orientation j");
        orientation_names.push_back("orientation k");
      orientation_=fh.get_keys<FloatTraits>(cat,
        orientation_names);
        Strings  coordinates_names;
        coordinates_names.push_back("cartesian x");
        coordinates_names.push_back("cartesian y");
        coordinates_names.push_back("cartesian z");
      coordinates_=fh.get_keys<FloatTraits>(cat,
        coordinates_names);;
    }
    /** Get a RigidParticleConst for nh.*/
    RigidParticleConst get(NodeConstHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  RigidParticle");;
      return RigidParticleConst(nh, orientation_,
coordinates_);
    }
    /** Check whether nh has all the attributes required to be a
        RigidParticleConst.*/
    bool get_is(NodeConstHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(orientation_[0])
    && nh.get_has_value(coordinates_[0]);
    }
    };

    typedef std::vector<RigidParticleConstFactory>
            RigidParticleConstFactories;

/** Information regarding diffusion coefficients.

       See also Diffuser and DiffuserConstFactory.
     */
    class DiffuserConst:
        public Decorator<NodeConstHandle> {
    friend class DiffuserConstFactory;
    std::string get_name() const {return "Diffuser";}
    private:
    typedef Decorator<NodeConstHandle> P;
    FloatKey diffusion_coefficient_;
    DiffuserConst(NodeConstHandle nh,
                  FloatKey diffusion_coefficient):
       P(nh),
diffusion_coefficient_(diffusion_coefficient) {
    ;
    }
    public:
    /** The diffusion coefficient in A^2/fs. */
Float get_diffusion_coefficient() const {
  try{
  return P::get_value(diffusion_coefficient_);
  } RMF_DECORATOR_CATCH( );
}
/** The diffusion coefficient in A^2/fs. */
Floats get_all_diffusion_coefficients() const {
  try {
  return P::get_all_values(diffusion_coefficient_);
  } RMF_DECORATOR_CATCH( );
}
    static std::string get_decorator_type_name() {
         return "DiffuserConst";
    }
    };

    typedef std::vector<DiffuserConst>
            DiffuserConsts;

/** Information regarding diffusion coefficients.

       See also DiffuserConst and DiffuserFactory.
     */
    class Diffuser:
        public Decorator<NodeHandle> {
    friend class DiffuserFactory;
    std::string get_name() const {return "Diffuser";}
    private:
    typedef Decorator<NodeHandle> P;
    FloatKey diffusion_coefficient_;
    Diffuser(NodeHandle nh,
                  FloatKey diffusion_coefficient):
       P(nh),
diffusion_coefficient_(diffusion_coefficient) {
    ;
    }
    public:
    /** The diffusion coefficient in A^2/fs. */
Float get_diffusion_coefficient() const {
  try{
  return P::get_value(diffusion_coefficient_);
  } RMF_DECORATOR_CATCH( );
}
/** The diffusion coefficient in A^2/fs. */
Floats get_all_diffusion_coefficients() const {
  try {
  return P::get_all_values(diffusion_coefficient_);
  } RMF_DECORATOR_CATCH( );
}
/** The diffusion coefficient in A^2/fs. */
void set_diffusion_coefficient(Float v) {
  try {
  P::set_value(diffusion_coefficient_,
            v);
  } RMF_DECORATOR_CATCH( );
}
    static std::string get_decorator_type_name() {
         return "Diffuser";
    }
    };

    typedef std::vector<Diffuser>
            Diffusers;

/** Create decorators of type Diffuser.

       See also Diffuser and DiffuserConstFactory.
    */
    class DiffuserFactory:
       public Factory<FileHandle>
                 {
    private:
    typedef Factory<FileHandle> P;
    FloatKey diffusion_coefficient_;
    public:
    DiffuserFactory(FileHandle fh):
      P(){
    Category cat = fh.get_category("physics");
RMF_UNUSED(cat);
diffusion_coefficient_=fh.get_key<FloatTraits>(cat,
        "diffusion coefficient");;
    }
    /** Get a Diffuser for nh.*/
    Diffuser get(NodeHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Diffuser");;
      return Diffuser(nh, diffusion_coefficient_);
    }
    /** Check whether nh has all the attributes required to be a
        Diffuser.*/
    bool get_is(NodeHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(diffusion_coefficient_);
    }
    };

    typedef std::vector<DiffuserFactory>
            DiffuserFactories;

/** Create decorators of type Diffuser.

       See also DiffuserConst and DiffuserFactory.
    */
    class DiffuserConstFactory:
       public Factory<FileConstHandle>
                 {
    private:
    typedef Factory<FileConstHandle> P;
    FloatKey diffusion_coefficient_;
    public:
    DiffuserConstFactory(FileConstHandle fh):
      P(){
    Category cat = fh.get_category("physics");
RMF_UNUSED(cat);
diffusion_coefficient_=fh.get_key<FloatTraits>(cat,
        "diffusion coefficient");;
    }
    /** Get a DiffuserConst for nh.*/
    DiffuserConst get(NodeConstHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Diffuser");;
      return DiffuserConst(nh, diffusion_coefficient_);
    }
    /** Check whether nh has all the attributes required to be a
        DiffuserConst.*/
    bool get_is(NodeConstHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(diffusion_coefficient_);
    }
    };

    typedef std::vector<DiffuserConstFactory>
            DiffuserConstFactories;

/** Information regarding an atom.

       See also Atom and AtomConstFactory.
     */
    class AtomConst:
        public Decorator<NodeConstHandle> {
    friend class AtomConstFactory;
    std::string get_name() const {return "Atom";}
    private:
    typedef Decorator<NodeConstHandle> P;
    FloatKeys coordinates_;
FloatKey radius_;
FloatKey mass_;
IndexKey element_;
    AtomConst(NodeConstHandle nh,
                  FloatKeys coordinates,
FloatKey radius,
FloatKey mass,
IndexKey element):
       P(nh),
coordinates_(coordinates),
radius_(radius),
mass_(mass),
element_(element) {
    ;
    }
    public:
    /** The coordinates in angstroms. */
Floats get_coordinates() const {
             return P::get_values(coordinates_);
           }
/** The radius in angstroms. */
Float get_radius() const {
  try{
  return P::get_value(radius_);
  } RMF_DECORATOR_CATCH( );
}
/** The radius in angstroms. */
Floats get_all_radiuss() const {
  try {
  return P::get_all_values(radius_);
  } RMF_DECORATOR_CATCH( );
}
/** The mass in Daltons. */
Float get_mass() const {
  try{
  return P::get_value(mass_);
  } RMF_DECORATOR_CATCH( );
}
/** The mass in Daltons. */
Floats get_all_masss() const {
  try {
  return P::get_all_values(mass_);
  } RMF_DECORATOR_CATCH( );
}
/** The atomic number of the element. */
Index get_element() const {
  try{
  return P::get_value(element_);
  } RMF_DECORATOR_CATCH( );
}
/** The atomic number of the element. */
Indexes get_all_elements() const {
  try {
  return P::get_all_values(element_);
  } RMF_DECORATOR_CATCH( );
}
    static std::string get_decorator_type_name() {
         return "AtomConst";
    }
    };

    typedef std::vector<AtomConst>
            AtomConsts;

/** Information regarding an atom.

       See also AtomConst and AtomFactory.
     */
    class Atom:
        public Decorator<NodeHandle> {
    friend class AtomFactory;
    std::string get_name() const {return "Atom";}
    private:
    typedef Decorator<NodeHandle> P;
    FloatKeys coordinates_;
FloatKey radius_;
FloatKey mass_;
IndexKey element_;
    Atom(NodeHandle nh,
                  FloatKeys coordinates,
FloatKey radius,
FloatKey mass,
IndexKey element):
       P(nh),
coordinates_(coordinates),
radius_(radius),
mass_(mass),
element_(element) {
    ;
    }
    public:
    /** The coordinates in angstroms. */
Floats get_coordinates() const {
             return P::get_values(coordinates_);
           }
/** The coordinates in angstroms. */
    void set_coordinates(const Floats &v) {
           P::set_values(coordinates_, v);
        }
/** The radius in angstroms. */
Float get_radius() const {
  try{
  return P::get_value(radius_);
  } RMF_DECORATOR_CATCH( );
}
/** The radius in angstroms. */
Floats get_all_radiuss() const {
  try {
  return P::get_all_values(radius_);
  } RMF_DECORATOR_CATCH( );
}
/** The radius in angstroms. */
void set_radius(Float v) {
  try {
  P::set_value(radius_,
            v);
  } RMF_DECORATOR_CATCH( );
}
/** The mass in Daltons. */
Float get_mass() const {
  try{
  return P::get_value(mass_);
  } RMF_DECORATOR_CATCH( );
}
/** The mass in Daltons. */
Floats get_all_masss() const {
  try {
  return P::get_all_values(mass_);
  } RMF_DECORATOR_CATCH( );
}
/** The mass in Daltons. */
void set_mass(Float v) {
  try {
  P::set_value(mass_,
            v);
  } RMF_DECORATOR_CATCH( );
}
/** The atomic number of the element. */
Index get_element() const {
  try{
  return P::get_value(element_);
  } RMF_DECORATOR_CATCH( );
}
/** The atomic number of the element. */
Indexes get_all_elements() const {
  try {
  return P::get_all_values(element_);
  } RMF_DECORATOR_CATCH( );
}
/** The atomic number of the element. */
void set_element(Index v) {
  try {
  P::set_value(element_,
            v);
  } RMF_DECORATOR_CATCH( );
}
    static std::string get_decorator_type_name() {
         return "Atom";
    }
    };

    typedef std::vector<Atom>
            Atoms;

/** Create decorators of type Atom.

       See also Atom and AtomConstFactory.
    */
    class AtomFactory:
       public Factory<FileHandle>
                 {
    private:
    typedef Factory<FileHandle> P;
    FloatKeys coordinates_;
FloatKey radius_;
FloatKey mass_;
IndexKey element_;
    public:
    AtomFactory(FileHandle fh):
      P(){
    Category cat = fh.get_category("physics");
RMF_UNUSED(cat);
        Strings  coordinates_names;
        coordinates_names.push_back("cartesian x");
        coordinates_names.push_back("cartesian y");
        coordinates_names.push_back("cartesian z");
      coordinates_=fh.get_keys<FloatTraits>(cat,
        coordinates_names);
radius_=fh.get_key<FloatTraits>(cat,
        "radius");
mass_=fh.get_key<FloatTraits>(cat,
        "mass");
element_=fh.get_key<IndexTraits>(cat,
        "element");;
    }
    /** Get a Atom for nh.*/
    Atom get(NodeHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Atom");;
      return Atom(nh, coordinates_,
radius_,
mass_,
element_);
    }
    /** Check whether nh has all the attributes required to be a
        Atom.*/
    bool get_is(NodeHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(coordinates_[0])
    && nh.get_has_value(radius_)
    && nh.get_has_value(mass_)
    && nh.get_has_value(element_);
    }
    };

    typedef std::vector<AtomFactory>
            AtomFactories;

/** Create decorators of type Atom.

       See also AtomConst and AtomFactory.
    */
    class AtomConstFactory:
       public Factory<FileConstHandle>
                 {
    private:
    typedef Factory<FileConstHandle> P;
    FloatKeys coordinates_;
FloatKey radius_;
FloatKey mass_;
IndexKey element_;
    public:
    AtomConstFactory(FileConstHandle fh):
      P(){
    Category cat = fh.get_category("physics");
RMF_UNUSED(cat);
        Strings  coordinates_names;
        coordinates_names.push_back("cartesian x");
        coordinates_names.push_back("cartesian y");
        coordinates_names.push_back("cartesian z");
      coordinates_=fh.get_keys<FloatTraits>(cat,
        coordinates_names);
radius_=fh.get_key<FloatTraits>(cat,
        "radius");
mass_=fh.get_key<FloatTraits>(cat,
        "mass");
element_=fh.get_key<IndexTraits>(cat,
        "element");;
    }
    /** Get a AtomConst for nh.*/
    AtomConst get(NodeConstHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Atom");;
      return AtomConst(nh, coordinates_,
radius_,
mass_,
element_);
    }
    /** Check whether nh has all the attributes required to be a
        AtomConst.*/
    bool get_is(NodeConstHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(coordinates_[0])
    && nh.get_has_value(radius_)
    && nh.get_has_value(mass_)
    && nh.get_has_value(element_);
    }
    };

    typedef std::vector<AtomConstFactory>
            AtomConstFactories;

/** A bond between particles.

       See also Bond and BondConstFactory.
     */
    class BondConst:
        public Decorator<NodeConstHandle> {
    friend class BondConstFactory;
    std::string get_name() const {return "Bond";}
    private:
    typedef Decorator<NodeConstHandle> P;
    AliasConstFactory bonded_;
    BondConst(NodeConstHandle nh,
                  AliasConstFactory bonded):
       P(nh),
bonded_(bonded) {
    ;
    }
    public:
    /** The bonded particles. */
NodeConstHandles get_bonded() const {
  try{
  NodeConstHandles typed=get_node().get_children();
  NodeConstHandles ret;
  for (unsigned int i=0; i< typed.size(); ++i) {
     if (bonded_.get_is(typed[i])) {
        ret.push_back(bonded_.get(typed[i]).get_aliased());
     }
  }
  return ret;
  } RMF_DECORATOR_CATCH( );
}
    static std::string get_decorator_type_name() {
         return "BondConst";
    }
    };

    typedef std::vector<BondConst>
            BondConsts;

/** A bond between particles.

       See also BondConst and BondFactory.
     */
    class Bond:
        public Decorator<NodeHandle> {
    friend class BondFactory;
    std::string get_name() const {return "Bond";}
    private:
    typedef Decorator<NodeHandle> P;
    AliasFactory bonded_;
    Bond(NodeHandle nh,
                  AliasFactory bonded):
       P(nh),
bonded_(bonded) {
    ;
    }
    public:
    /** The bonded particles. */
NodeHandles get_bonded() const {
  try{
  NodeHandles typed=get_node().get_children();
  NodeHandles ret;
  for (unsigned int i=0; i< typed.size(); ++i) {
     if (bonded_.get_is(typed[i])) {
        ret.push_back(bonded_.get(typed[i]).get_aliased());
     }
  }
  return ret;
  } RMF_DECORATOR_CATCH( );
}
/** The bonded particles. */
void set_bonded(NodeConstHandles v) {
  try{
   for (unsigned int i=0; i< v.size(); ++i) {
       internal::add_child_alias(bonded_, get_node(), v[i]);
   }
  } RMF_DECORATOR_CATCH( );
}
/** The bonded particles. */
void set_bonded(NodeHandles v) {
  try{
   for (unsigned int i=0; i< v.size(); ++i) {
       internal::add_child_alias(bonded_, get_node(), v[i]);
   }
  } RMF_DECORATOR_CATCH( );
}
    static std::string get_decorator_type_name() {
         return "Bond";
    }
    };

    typedef std::vector<Bond>
            Bonds;

/** Create decorators of type Bond.

       See also Bond and BondConstFactory.
    */
    class BondFactory:
       public Factory<FileHandle>
                 {
    private:
    typedef Factory<FileHandle> P;
    AliasFactory bonded_;
    public:
    BondFactory(FileHandle fh):
      P(), bonded_(fh){
    Category cat = fh.get_category("physics");
RMF_UNUSED(cat);;
    }
    /** Get a Bond for nh.*/
    Bond get(NodeHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::BOND), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Bond");;
      return Bond(nh, bonded_);
    }
    /** Check whether nh has all the attributes required to be a
        Bond.*/
    bool get_is(NodeHandle nh) const {
      return (nh.get_type() == RMF::BOND);
    }
    };

    typedef std::vector<BondFactory>
            BondFactories;

/** Create decorators of type Bond.

       See also BondConst and BondFactory.
    */
    class BondConstFactory:
       public Factory<FileConstHandle>
                 {
    private:
    typedef Factory<FileConstHandle> P;
    AliasConstFactory bonded_;
    public:
    BondConstFactory(FileConstHandle fh):
      P(), bonded_(fh){
    Category cat = fh.get_category("physics");
RMF_UNUSED(cat);;
    }
    /** Get a BondConst for nh.*/
    BondConst get(NodeConstHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::BOND), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Bond");;
      return BondConst(nh, bonded_);
    }
    /** Check whether nh has all the attributes required to be a
        BondConst.*/
    bool get_is(NodeConstHandle nh) const {
      return (nh.get_type() == RMF::BOND);
    }
    };

    typedef std::vector<BondConstFactory>
            BondConstFactories;

/** Define a transformation to be applied the the attributes of this and child particles, relative to the reference frame of the parent.

       See also ReferenceFrame and ReferenceFrameConstFactory.
     */
    class ReferenceFrameConst:
        public Decorator<NodeConstHandle> {
    friend class ReferenceFrameConstFactory;
    std::string get_name() const {return "ReferenceFrame";}
    private:
    typedef Decorator<NodeConstHandle> P;
    FloatKeys rotation_;
FloatKeys translation_;
    ReferenceFrameConst(NodeConstHandle nh,
                  FloatKeys rotation,
FloatKeys translation):
       P(nh),
rotation_(rotation),
translation_(translation) {
    ;
    }
    public:
    /** The rotational part of the relative transformation as a quaternion. */
Floats get_rotation() const {
             return P::get_values(rotation_);
           }
/** The translation part of the relative transformion in angstroms. */
Floats get_translation() const {
             return P::get_values(translation_);
           }
    static std::string get_decorator_type_name() {
         return "ReferenceFrameConst";
    }
    };

    typedef std::vector<ReferenceFrameConst>
            ReferenceFrameConsts;

/** Define a transformation to be applied the the attributes of this and child particles, relative to the reference frame of the parent.

       See also ReferenceFrameConst and ReferenceFrameFactory.
     */
    class ReferenceFrame:
        public Decorator<NodeHandle> {
    friend class ReferenceFrameFactory;
    std::string get_name() const {return "ReferenceFrame";}
    private:
    typedef Decorator<NodeHandle> P;
    FloatKeys rotation_;
FloatKeys translation_;
    ReferenceFrame(NodeHandle nh,
                  FloatKeys rotation,
FloatKeys translation):
       P(nh),
rotation_(rotation),
translation_(translation) {
    ;
    }
    public:
    /** The rotational part of the relative transformation as a quaternion. */
Floats get_rotation() const {
             return P::get_values(rotation_);
           }
/** The rotational part of the relative transformation as a quaternion. */
    void set_rotation(const Floats &v) {
           P::set_values(rotation_, v);
        }
/** The translation part of the relative transformion in angstroms. */
Floats get_translation() const {
             return P::get_values(translation_);
           }
/** The translation part of the relative transformion in angstroms. */
    void set_translation(const Floats &v) {
           P::set_values(translation_, v);
        }
    static std::string get_decorator_type_name() {
         return "ReferenceFrame";
    }
    };

    typedef std::vector<ReferenceFrame>
            ReferenceFrames;

/** Create decorators of type ReferenceFrame.

       See also ReferenceFrame and ReferenceFrameConstFactory.
    */
    class ReferenceFrameFactory:
       public Factory<FileHandle>
                 {
    private:
    typedef Factory<FileHandle> P;
    FloatKeys rotation_;
FloatKeys translation_;
    public:
    ReferenceFrameFactory(FileHandle fh):
      P(){
    Category cat = fh.get_category("physics");
RMF_UNUSED(cat);
        Strings  rotation_names;
        rotation_names.push_back("reference frame orientation r");
        rotation_names.push_back("reference frame orientation i");
        rotation_names.push_back("reference frame orientation j");
        rotation_names.push_back("reference frame orientation k");
      rotation_=fh.get_keys<FloatTraits>(cat,
        rotation_names);
        Strings  translation_names;
        translation_names.push_back("reference frame cartesian x");
        translation_names.push_back("reference frame cartesian y");
        translation_names.push_back("reference frame cartesian z");
      translation_=fh.get_keys<FloatTraits>(cat,
        translation_names);;
    }
    /** Get a ReferenceFrame for nh.*/
    ReferenceFrame get(NodeHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION||nh.get_type() == RMF::ORGANIZATIONAL), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  ReferenceFrame");;
      return ReferenceFrame(nh, rotation_,
translation_);
    }
    /** Check whether nh has all the attributes required to be a
        ReferenceFrame.*/
    bool get_is(NodeHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION||nh.get_type() == RMF::ORGANIZATIONAL)
    && nh.get_has_value(rotation_[0])
    && nh.get_has_value(translation_[0]);
    }
    };

    typedef std::vector<ReferenceFrameFactory>
            ReferenceFrameFactories;

/** Create decorators of type ReferenceFrame.

       See also ReferenceFrameConst and ReferenceFrameFactory.
    */
    class ReferenceFrameConstFactory:
       public Factory<FileConstHandle>
                 {
    private:
    typedef Factory<FileConstHandle> P;
    FloatKeys rotation_;
FloatKeys translation_;
    public:
    ReferenceFrameConstFactory(FileConstHandle fh):
      P(){
    Category cat = fh.get_category("physics");
RMF_UNUSED(cat);
        Strings  rotation_names;
        rotation_names.push_back("reference frame orientation r");
        rotation_names.push_back("reference frame orientation i");
        rotation_names.push_back("reference frame orientation j");
        rotation_names.push_back("reference frame orientation k");
      rotation_=fh.get_keys<FloatTraits>(cat,
        rotation_names);
        Strings  translation_names;
        translation_names.push_back("reference frame cartesian x");
        translation_names.push_back("reference frame cartesian y");
        translation_names.push_back("reference frame cartesian z");
      translation_=fh.get_keys<FloatTraits>(cat,
        translation_names);;
    }
    /** Get a ReferenceFrameConst for nh.*/
    ReferenceFrameConst get(NodeConstHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION||nh.get_type() == RMF::ORGANIZATIONAL), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  ReferenceFrame");;
      return ReferenceFrameConst(nh, rotation_,
translation_);
    }
    /** Check whether nh has all the attributes required to be a
        ReferenceFrameConst.*/
    bool get_is(NodeConstHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION||nh.get_type() == RMF::ORGANIZATIONAL)
    && nh.get_has_value(rotation_[0])
    && nh.get_has_value(translation_[0]);
    }
    };

    typedef std::vector<ReferenceFrameConstFactory>
            ReferenceFrameConstFactories;

/** Forces acting on particles in kCal/mol/A.

       See also Force and ForceConstFactory.
     */
    class ForceConst:
        public Decorator<NodeConstHandle> {
    friend class ForceConstFactory;
    std::string get_name() const {return "Force";}
    private:
    typedef Decorator<NodeConstHandle> P;
    FloatKeys force_;
    ForceConst(NodeConstHandle nh,
                  FloatKeys force):
       P(nh),
force_(force) {
    ;
    }
    public:
    /** The force. */
Floats get_force() const {
             return P::get_values(force_);
           }
    static std::string get_decorator_type_name() {
         return "ForceConst";
    }
    };

    typedef std::vector<ForceConst>
            ForceConsts;

/** Forces acting on particles in kCal/mol/A.

       See also ForceConst and ForceFactory.
     */
    class Force:
        public Decorator<NodeHandle> {
    friend class ForceFactory;
    std::string get_name() const {return "Force";}
    private:
    typedef Decorator<NodeHandle> P;
    FloatKeys force_;
    Force(NodeHandle nh,
                  FloatKeys force):
       P(nh),
force_(force) {
    ;
    }
    public:
    /** The force. */
Floats get_force() const {
             return P::get_values(force_);
           }
/** The force. */
    void set_force(const Floats &v) {
           P::set_values(force_, v);
        }
    static std::string get_decorator_type_name() {
         return "Force";
    }
    };

    typedef std::vector<Force>
            Forces;

/** Create decorators of type Force.

       See also Force and ForceConstFactory.
    */
    class ForceFactory:
       public Factory<FileHandle>
                 {
    private:
    typedef Factory<FileHandle> P;
    FloatKeys force_;
    public:
    ForceFactory(FileHandle fh):
      P(){
    Category cat = fh.get_category("physics");
RMF_UNUSED(cat);
        Strings  force_names;
        force_names.push_back("force cartesian x");
        force_names.push_back("force cartesian y");
        force_names.push_back("force cartesian z");
      force_=fh.get_keys<FloatTraits>(cat,
        force_names);;
    }
    /** Get a Force for nh.*/
    Force get(NodeHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Force");;
      return Force(nh, force_);
    }
    /** Check whether nh has all the attributes required to be a
        Force.*/
    bool get_is(NodeHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(force_[0]);
    }
    };

    typedef std::vector<ForceFactory>
            ForceFactories;

/** Create decorators of type Force.

       See also ForceConst and ForceFactory.
    */
    class ForceConstFactory:
       public Factory<FileConstHandle>
                 {
    private:
    typedef Factory<FileConstHandle> P;
    FloatKeys force_;
    public:
    ForceConstFactory(FileConstHandle fh):
      P(){
    Category cat = fh.get_category("physics");
RMF_UNUSED(cat);
        Strings  force_names;
        force_names.push_back("force cartesian x");
        force_names.push_back("force cartesian y");
        force_names.push_back("force cartesian z");
      force_=fh.get_keys<FloatTraits>(cat,
        force_names);;
    }
    /** Get a ForceConst for nh.*/
    ForceConst get(NodeConstHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Force");;
      return ForceConst(nh, force_);
    }
    /** Check whether nh has all the attributes required to be a
        ForceConst.*/
    bool get_is(NodeConstHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(force_[0]);
    }
    };

    typedef std::vector<ForceConstFactory>
            ForceConstFactories;

/** Torque acting on particles in kCal/mol/radian.

       See also Torque and TorqueConstFactory.
     */
    class TorqueConst:
        public Decorator<NodeConstHandle> {
    friend class TorqueConstFactory;
    std::string get_name() const {return "Torque";}
    private:
    typedef Decorator<NodeConstHandle> P;
    FloatKeys torque_;
    TorqueConst(NodeConstHandle nh,
                  FloatKeys torque):
       P(nh),
torque_(torque) {
    ;
    }
    public:
    /** The torque. */
Floats get_torque() const {
             return P::get_values(torque_);
           }
    static std::string get_decorator_type_name() {
         return "TorqueConst";
    }
    };

    typedef std::vector<TorqueConst>
            TorqueConsts;

/** Torque acting on particles in kCal/mol/radian.

       See also TorqueConst and TorqueFactory.
     */
    class Torque:
        public Decorator<NodeHandle> {
    friend class TorqueFactory;
    std::string get_name() const {return "Torque";}
    private:
    typedef Decorator<NodeHandle> P;
    FloatKeys torque_;
    Torque(NodeHandle nh,
                  FloatKeys torque):
       P(nh),
torque_(torque) {
    ;
    }
    public:
    /** The torque. */
Floats get_torque() const {
             return P::get_values(torque_);
           }
/** The torque. */
    void set_torque(const Floats &v) {
           P::set_values(torque_, v);
        }
    static std::string get_decorator_type_name() {
         return "Torque";
    }
    };

    typedef std::vector<Torque>
            Torques;

/** Create decorators of type Torque.

       See also Torque and TorqueConstFactory.
    */
    class TorqueFactory:
       public Factory<FileHandle>
                 {
    private:
    typedef Factory<FileHandle> P;
    FloatKeys torque_;
    public:
    TorqueFactory(FileHandle fh):
      P(){
    Category cat = fh.get_category("physics");
RMF_UNUSED(cat);
        Strings  torque_names;
        torque_names.push_back("torque cartesian x");
        torque_names.push_back("torque cartesian y");
        torque_names.push_back("torque cartesian z");
      torque_=fh.get_keys<FloatTraits>(cat,
        torque_names);;
    }
    /** Get a Torque for nh.*/
    Torque get(NodeHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Torque");;
      return Torque(nh, torque_);
    }
    /** Check whether nh has all the attributes required to be a
        Torque.*/
    bool get_is(NodeHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(torque_[0]);
    }
    };

    typedef std::vector<TorqueFactory>
            TorqueFactories;

/** Create decorators of type Torque.

       See also TorqueConst and TorqueFactory.
    */
    class TorqueConstFactory:
       public Factory<FileConstHandle>
                 {
    private:
    typedef Factory<FileConstHandle> P;
    FloatKeys torque_;
    public:
    TorqueConstFactory(FileConstHandle fh):
      P(){
    Category cat = fh.get_category("physics");
RMF_UNUSED(cat);
        Strings  torque_names;
        torque_names.push_back("torque cartesian x");
        torque_names.push_back("torque cartesian y");
        torque_names.push_back("torque cartesian z");
      torque_=fh.get_keys<FloatTraits>(cat,
        torque_names);;
    }
    /** Get a TorqueConst for nh.*/
    TorqueConst get(NodeConstHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Torque");;
      return TorqueConst(nh, torque_);
    }
    /** Check whether nh has all the attributes required to be a
        TorqueConst.*/
    bool get_is(NodeConstHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(torque_[0]);
    }
    };

    typedef std::vector<TorqueConstFactory>
            TorqueConstFactories;

} /* namespace RMF */
RMF_DISABLE_WARNINGS

#endif /* RMF_PHYSICS_DECORATORS_H */
