/**
 *  \file RMF/sequence_decorators.h
 *  \brief Helper functions for manipulating RMF files.
 *
 *  Copyright 2007-2013 IMP Inventors. All rights reserved.
 *
 */

#ifndef RMF_SEQUENCE_DECORATORS_H
#define RMF_SEQUENCE_DECORATORS_H

#include <RMF/config.h>
#include <RMF/infrastructure_macros.h>
#include <RMF/NodeHandle.h>
#include <RMF/FileHandle.h>
#include <RMF/Decorator.h>
#include <RMF/Factory.h>
#include <RMF/constants.h>
#include <RMF/internal/utility.h>
#include <RMF/internal/paths.h>
#include <boost/array.hpp>
#include <boost/lexical_cast.hpp>

RMF_ENABLE_WARNINGS
namespace RMF {

/** Information regarding a residue.

       See also Residue and ResidueConstFactory.
     */
    class ResidueConst:
        public Decorator<NodeConstHandle> {
    friend class ResidueConstFactory;
    std::string get_name() const {return "Residue";}
    private:
    typedef Decorator<NodeConstHandle> P;
    boost::array<IntKey,2> index_;
StringKey type_;
    ResidueConst(NodeConstHandle nh,
                  boost::array<IntKey, 2> index,
StringKey type):
       P(nh),
index_(index),
type_(type) {
    ;
    }
    public:
    /** The index of the residue. */
Int get_index() const {
  try {
  return get_node().get_value(index_[0]);
  } RMF_DECORATOR_CATCH( );
}
/** The three letter name for the residue. */
String get_type() const {
  try{
  return P::get_value(type_);
  } RMF_DECORATOR_CATCH( );
}
/** The three letter name for the residue. */
Strings get_all_types() const {
  try {
  return P::get_all_values(type_);
  } RMF_DECORATOR_CATCH( );
}
    static std::string get_decorator_type_name() {
         return "ResidueConst";
    }
    };

    typedef std::vector<ResidueConst>
            ResidueConsts;

/** Information regarding a residue.

       See also ResidueConst and ResidueFactory.
     */
    class Residue:
        public Decorator<NodeHandle> {
    friend class ResidueFactory;
    std::string get_name() const {return "Residue";}
    private:
    typedef Decorator<NodeHandle> P;
    boost::array<IntKey,2> index_;
StringKey type_;
    Residue(NodeHandle nh,
                  boost::array<IntKey, 2> index,
StringKey type):
       P(nh),
index_(index),
type_(type) {
    ;
    }
    public:
    /** The index of the residue. */
Int get_index() const {
  try {
  return get_node().get_value(index_[0]);
  } RMF_DECORATOR_CATCH( );
}
/** The index of the residue. */
void set_index(Int v) {
  try {
   get_node().set_value(index_[0], v);
   get_node().set_value(index_[1], v);
  } RMF_DECORATOR_CATCH( );
}
/** The three letter name for the residue. */
String get_type() const {
  try{
  return P::get_value(type_);
  } RMF_DECORATOR_CATCH( );
}
/** The three letter name for the residue. */
Strings get_all_types() const {
  try {
  return P::get_all_values(type_);
  } RMF_DECORATOR_CATCH( );
}
/** The three letter name for the residue. */
void set_type(String v) {
  try {
  P::set_value(type_,
            v);
  } RMF_DECORATOR_CATCH( );
}
    static std::string get_decorator_type_name() {
         return "Residue";
    }
    };

    typedef std::vector<Residue>
            Residues;

/** Create decorators of type Residue.

       See also Residue and ResidueConstFactory.
    */
    class ResidueFactory:
       public Factory<FileHandle>
                 {
    private:
    typedef Factory<FileHandle> P;
    boost::array<IntKey,2> index_;
StringKey type_;
    public:
    ResidueFactory(FileHandle fh):
      P(){
    Category cat = fh.get_category("sequence");
RMF_UNUSED(cat);
index_[0]=fh.get_key<IntTraits>(cat,
        "first residue index");
index_[1]=fh.get_key<IntTraits>(cat,
        "last residue index");
type_=fh.get_key<StringTraits>(cat,
        "residue type");;
    }
    /** Get a Residue for nh.*/
    Residue get(NodeHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Residue");;
      return Residue(nh, index_,
type_);
    }
    /** Check whether nh has all the attributes required to be a
        Residue.*/
    bool get_is(NodeHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(index_[0])
  && nh.get_has_value(index_[1])
  && nh.get_value(index_[0])
   ==nh.get_value(index_[1])
    && nh.get_has_value(type_);
    }
    };

    typedef std::vector<ResidueFactory>
            ResidueFactories;

/** Create decorators of type Residue.

       See also ResidueConst and ResidueFactory.
    */
    class ResidueConstFactory:
       public Factory<FileConstHandle>
                 {
    private:
    typedef Factory<FileConstHandle> P;
    boost::array<IntKey,2> index_;
StringKey type_;
    public:
    ResidueConstFactory(FileConstHandle fh):
      P(){
    Category cat = fh.get_category("sequence");
RMF_UNUSED(cat);
index_[0]=fh.get_key<IntTraits>(cat,
        "first residue index");
index_[1]=fh.get_key<IntTraits>(cat,
        "last residue index");
type_=fh.get_key<StringTraits>(cat,
        "residue type");;
    }
    /** Get a ResidueConst for nh.*/
    ResidueConst get(NodeConstHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Residue");;
      return ResidueConst(nh, index_,
type_);
    }
    /** Check whether nh has all the attributes required to be a
        ResidueConst.*/
    bool get_is(NodeConstHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(index_[0])
  && nh.get_has_value(index_[1])
  && nh.get_value(index_[0])
   ==nh.get_value(index_[1])
    && nh.get_has_value(type_);
    }
    };

    typedef std::vector<ResidueConstFactory>
            ResidueConstFactories;

/** Information regarding a chain.

       See also Chain and ChainConstFactory.
     */
    class ChainConst:
        public Decorator<NodeConstHandle> {
    friend class ChainConstFactory;
    std::string get_name() const {return "Chain";}
    private:
    typedef Decorator<NodeConstHandle> P;
    IndexKey chain_id_;
    ChainConst(NodeConstHandle nh,
                  IndexKey chain_id):
       P(nh),
chain_id_(chain_id) {
    ;
    }
    public:
    /** The one letter id for the chain (A is 0 etc). */
Index get_chain_id() const {
  try{
  return P::get_value(chain_id_);
  } RMF_DECORATOR_CATCH( );
}
/** The one letter id for the chain (A is 0 etc). */
Indexes get_all_chain_ids() const {
  try {
  return P::get_all_values(chain_id_);
  } RMF_DECORATOR_CATCH( );
}
    static std::string get_decorator_type_name() {
         return "ChainConst";
    }
    };

    typedef std::vector<ChainConst>
            ChainConsts;

/** Information regarding a chain.

       See also ChainConst and ChainFactory.
     */
    class Chain:
        public Decorator<NodeHandle> {
    friend class ChainFactory;
    std::string get_name() const {return "Chain";}
    private:
    typedef Decorator<NodeHandle> P;
    IndexKey chain_id_;
    Chain(NodeHandle nh,
                  IndexKey chain_id):
       P(nh),
chain_id_(chain_id) {
    ;
    }
    public:
    /** The one letter id for the chain (A is 0 etc). */
Index get_chain_id() const {
  try{
  return P::get_value(chain_id_);
  } RMF_DECORATOR_CATCH( );
}
/** The one letter id for the chain (A is 0 etc). */
Indexes get_all_chain_ids() const {
  try {
  return P::get_all_values(chain_id_);
  } RMF_DECORATOR_CATCH( );
}
/** The one letter id for the chain (A is 0 etc). */
void set_chain_id(Index v) {
  try {
  P::set_value(chain_id_,
            v);
  } RMF_DECORATOR_CATCH( );
}
    static std::string get_decorator_type_name() {
         return "Chain";
    }
    };

    typedef std::vector<Chain>
            Chains;

/** Create decorators of type Chain.

       See also Chain and ChainConstFactory.
    */
    class ChainFactory:
       public Factory<FileHandle>
                 {
    private:
    typedef Factory<FileHandle> P;
    IndexKey chain_id_;
    public:
    ChainFactory(FileHandle fh):
      P(){
    Category cat = fh.get_category("sequence");
RMF_UNUSED(cat);
chain_id_=fh.get_key<IndexTraits>(cat,
        "chain id");;
    }
    /** Get a Chain for nh.*/
    Chain get(NodeHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Chain");;
      return Chain(nh, chain_id_);
    }
    /** Check whether nh has all the attributes required to be a
        Chain.*/
    bool get_is(NodeHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(chain_id_);
    }
    };

    typedef std::vector<ChainFactory>
            ChainFactories;

/** Create decorators of type Chain.

       See also ChainConst and ChainFactory.
    */
    class ChainConstFactory:
       public Factory<FileConstHandle>
                 {
    private:
    typedef Factory<FileConstHandle> P;
    IndexKey chain_id_;
    public:
    ChainConstFactory(FileConstHandle fh):
      P(){
    Category cat = fh.get_category("sequence");
RMF_UNUSED(cat);
chain_id_=fh.get_key<IndexTraits>(cat,
        "chain id");;
    }
    /** Get a ChainConst for nh.*/
    ChainConst get(NodeConstHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Chain");;
      return ChainConst(nh, chain_id_);
    }
    /** Check whether nh has all the attributes required to be a
        ChainConst.*/
    bool get_is(NodeConstHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(chain_id_);
    }
    };

    typedef std::vector<ChainConstFactory>
            ChainConstFactories;

/** Information regarding a fragment of a molecule.

       See also Domain and DomainConstFactory.
     */
    class DomainConst:
        public Decorator<NodeConstHandle> {
    friend class DomainConstFactory;
    std::string get_name() const {return "Domain";}
    private:
    typedef Decorator<NodeConstHandle> P;
    boost::array<IntKey,2> indexes_;
    DomainConst(NodeConstHandle nh,
                  boost::array<IntKey, 2> indexes):
       P(nh),
indexes_(indexes) {
    ;
    }
    public:
    /** The range for the residues, specified as [first_index...last_index]. */
IntRange get_indexes() const {
  try {
  return std::make_pair(get_node().get_value(indexes_[0]),
                        get_node().get_value(indexes_[1]));
  } RMF_DECORATOR_CATCH( );
}
    static std::string get_decorator_type_name() {
         return "DomainConst";
    }
    };

    typedef std::vector<DomainConst>
            DomainConsts;

/** Information regarding a fragment of a molecule.

       See also DomainConst and DomainFactory.
     */
    class Domain:
        public Decorator<NodeHandle> {
    friend class DomainFactory;
    std::string get_name() const {return "Domain";}
    private:
    typedef Decorator<NodeHandle> P;
    boost::array<IntKey,2> indexes_;
    Domain(NodeHandle nh,
                  boost::array<IntKey, 2> indexes):
       P(nh),
indexes_(indexes) {
    ;
    }
    public:
    /** The range for the residues, specified as [first_index...last_index]. */
IntRange get_indexes() const {
  try {
  return std::make_pair(get_node().get_value(indexes_[0]),
                        get_node().get_value(indexes_[1]));
  } RMF_DECORATOR_CATCH( );
}
/** The range for the residues, specified as [first_index...last_index]. */
void set_indexes(Int v0, Int v1) {
 try {
   get_node().set_value(indexes_[0], v0);
   get_node().set_value(indexes_[1], v1);
  } RMF_DECORATOR_CATCH( );
}
    static std::string get_decorator_type_name() {
         return "Domain";
    }
    };

    typedef std::vector<Domain>
            Domains;

/** Create decorators of type Domain.

       See also Domain and DomainConstFactory.
    */
    class DomainFactory:
       public Factory<FileHandle>
                 {
    private:
    typedef Factory<FileHandle> P;
    boost::array<IntKey,2> indexes_;
    public:
    DomainFactory(FileHandle fh):
      P(){
    Category cat = fh.get_category("sequence");
RMF_UNUSED(cat);
indexes_[0]=fh.get_key<IntTraits>(cat,
        "first residue index");
indexes_[1]=fh.get_key<IntTraits>(cat,
        "last residue index");;
    }
    /** Get a Domain for nh.*/
    Domain get(NodeHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Domain");;
      return Domain(nh, indexes_);
    }
    /** Check whether nh has all the attributes required to be a
        Domain.*/
    bool get_is(NodeHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(indexes_[0])
  && nh.get_has_value(indexes_[1])
  && nh.get_value(indexes_[0])
   <nh.get_value(indexes_[1]);
    }
    };

    typedef std::vector<DomainFactory>
            DomainFactories;

/** Create decorators of type Domain.

       See also DomainConst and DomainFactory.
    */
    class DomainConstFactory:
       public Factory<FileConstHandle>
                 {
    private:
    typedef Factory<FileConstHandle> P;
    boost::array<IntKey,2> indexes_;
    public:
    DomainConstFactory(FileConstHandle fh):
      P(){
    Category cat = fh.get_category("sequence");
RMF_UNUSED(cat);
indexes_[0]=fh.get_key<IntTraits>(cat,
        "first residue index");
indexes_[1]=fh.get_key<IntTraits>(cat,
        "last residue index");;
    }
    /** Get a DomainConst for nh.*/
    DomainConst get(NodeConstHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Domain");;
      return DomainConst(nh, indexes_);
    }
    /** Check whether nh has all the attributes required to be a
        DomainConst.*/
    bool get_is(NodeConstHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(indexes_[0])
  && nh.get_has_value(indexes_[1])
  && nh.get_value(indexes_[0])
   <nh.get_value(indexes_[1]);
    }
    };

    typedef std::vector<DomainConstFactory>
            DomainConstFactories;

/** A numeric tag for keeping track of types of molecules.

       See also Typed and TypedConstFactory.
     */
    class TypedConst:
        public Decorator<NodeConstHandle> {
    friend class TypedConstFactory;
    std::string get_name() const {return "Typed";}
    private:
    typedef Decorator<NodeConstHandle> P;
    StringKey type_name_;
    TypedConst(NodeConstHandle nh,
                  StringKey type_name):
       P(nh),
type_name_(type_name) {
    ;
    }
    public:
    /** An arbitrary tag representing the type. */
String get_type_name() const {
  try{
  return P::get_value(type_name_);
  } RMF_DECORATOR_CATCH( );
}
/** An arbitrary tag representing the type. */
Strings get_all_type_names() const {
  try {
  return P::get_all_values(type_name_);
  } RMF_DECORATOR_CATCH( );
}
    static std::string get_decorator_type_name() {
         return "TypedConst";
    }
    };

    typedef std::vector<TypedConst>
            TypedConsts;

/** A numeric tag for keeping track of types of molecules.

       See also TypedConst and TypedFactory.
     */
    class Typed:
        public Decorator<NodeHandle> {
    friend class TypedFactory;
    std::string get_name() const {return "Typed";}
    private:
    typedef Decorator<NodeHandle> P;
    StringKey type_name_;
    Typed(NodeHandle nh,
                  StringKey type_name):
       P(nh),
type_name_(type_name) {
    ;
    }
    public:
    /** An arbitrary tag representing the type. */
String get_type_name() const {
  try{
  return P::get_value(type_name_);
  } RMF_DECORATOR_CATCH( );
}
/** An arbitrary tag representing the type. */
Strings get_all_type_names() const {
  try {
  return P::get_all_values(type_name_);
  } RMF_DECORATOR_CATCH( );
}
/** An arbitrary tag representing the type. */
void set_type_name(String v) {
  try {
  P::set_value(type_name_,
            v);
  } RMF_DECORATOR_CATCH( );
}
    static std::string get_decorator_type_name() {
         return "Typed";
    }
    };

    typedef std::vector<Typed>
            Typeds;

/** Create decorators of type Typed.

       See also Typed and TypedConstFactory.
    */
    class TypedFactory:
       public Factory<FileHandle>
                 {
    private:
    typedef Factory<FileHandle> P;
    StringKey type_name_;
    public:
    TypedFactory(FileHandle fh):
      P(){
    Category cat = fh.get_category("sequence");
RMF_UNUSED(cat);
type_name_=fh.get_key<StringTraits>(cat,
        "type name");;
    }
    /** Get a Typed for nh.*/
    Typed get(NodeHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Typed");;
      return Typed(nh, type_name_);
    }
    /** Check whether nh has all the attributes required to be a
        Typed.*/
    bool get_is(NodeHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(type_name_);
    }
    };

    typedef std::vector<TypedFactory>
            TypedFactories;

/** Create decorators of type Typed.

       See also TypedConst and TypedFactory.
    */
    class TypedConstFactory:
       public Factory<FileConstHandle>
                 {
    private:
    typedef Factory<FileConstHandle> P;
    StringKey type_name_;
    public:
    TypedConstFactory(FileConstHandle fh):
      P(){
    Category cat = fh.get_category("sequence");
RMF_UNUSED(cat);
type_name_=fh.get_key<StringTraits>(cat,
        "type name");;
    }
    /** Get a TypedConst for nh.*/
    TypedConst get(NodeConstHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Typed");;
      return TypedConst(nh, type_name_);
    }
    /** Check whether nh has all the attributes required to be a
        TypedConst.*/
    bool get_is(NodeConstHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(type_name_);
    }
    };

    typedef std::vector<TypedConstFactory>
            TypedConstFactories;

/** Information regarding a copy of a molecule.

       See also Copy and CopyConstFactory.
     */
    class CopyConst:
        public Decorator<NodeConstHandle> {
    friend class CopyConstFactory;
    std::string get_name() const {return "Copy";}
    private:
    typedef Decorator<NodeConstHandle> P;
    IndexKey copy_index_;
    CopyConst(NodeConstHandle nh,
                  IndexKey copy_index):
       P(nh),
copy_index_(copy_index) {
    ;
    }
    public:
    /** This is the copy_indexth copy of the original. */
Index get_copy_index() const {
  try{
  return P::get_value(copy_index_);
  } RMF_DECORATOR_CATCH( );
}
/** This is the copy_indexth copy of the original. */
Indexes get_all_copy_indexs() const {
  try {
  return P::get_all_values(copy_index_);
  } RMF_DECORATOR_CATCH( );
}
    static std::string get_decorator_type_name() {
         return "CopyConst";
    }
    };

    typedef std::vector<CopyConst>
            CopyConsts;

/** Information regarding a copy of a molecule.

       See also CopyConst and CopyFactory.
     */
    class Copy:
        public Decorator<NodeHandle> {
    friend class CopyFactory;
    std::string get_name() const {return "Copy";}
    private:
    typedef Decorator<NodeHandle> P;
    IndexKey copy_index_;
    Copy(NodeHandle nh,
                  IndexKey copy_index):
       P(nh),
copy_index_(copy_index) {
    ;
    }
    public:
    /** This is the copy_indexth copy of the original. */
Index get_copy_index() const {
  try{
  return P::get_value(copy_index_);
  } RMF_DECORATOR_CATCH( );
}
/** This is the copy_indexth copy of the original. */
Indexes get_all_copy_indexs() const {
  try {
  return P::get_all_values(copy_index_);
  } RMF_DECORATOR_CATCH( );
}
/** This is the copy_indexth copy of the original. */
void set_copy_index(Index v) {
  try {
  P::set_value(copy_index_,
            v);
  } RMF_DECORATOR_CATCH( );
}
    static std::string get_decorator_type_name() {
         return "Copy";
    }
    };

    typedef std::vector<Copy>
            Copys;

/** Create decorators of type Copy.

       See also Copy and CopyConstFactory.
    */
    class CopyFactory:
       public Factory<FileHandle>
                 {
    private:
    typedef Factory<FileHandle> P;
    IndexKey copy_index_;
    public:
    CopyFactory(FileHandle fh):
      P(){
    Category cat = fh.get_category("sequence");
RMF_UNUSED(cat);
copy_index_=fh.get_key<IndexTraits>(cat,
        "copy index");;
    }
    /** Get a Copy for nh.*/
    Copy get(NodeHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Copy");;
      return Copy(nh, copy_index_);
    }
    /** Check whether nh has all the attributes required to be a
        Copy.*/
    bool get_is(NodeHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(copy_index_);
    }
    };

    typedef std::vector<CopyFactory>
            CopyFactories;

/** Create decorators of type Copy.

       See also CopyConst and CopyFactory.
    */
    class CopyConstFactory:
       public Factory<FileConstHandle>
                 {
    private:
    typedef Factory<FileConstHandle> P;
    IndexKey copy_index_;
    public:
    CopyConstFactory(FileConstHandle fh):
      P(){
    Category cat = fh.get_category("sequence");
RMF_UNUSED(cat);
copy_index_=fh.get_key<IndexTraits>(cat,
        "copy index");;
    }
    /** Get a CopyConst for nh.*/
    CopyConst get(NodeConstHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Copy");;
      return CopyConst(nh, copy_index_);
    }
    /** Check whether nh has all the attributes required to be a
        CopyConst.*/
    bool get_is(NodeConstHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION)
    && nh.get_has_value(copy_index_);
    }
    };

    typedef std::vector<CopyConstFactory>
            CopyConstFactories;

} /* namespace RMF */
RMF_DISABLE_WARNINGS

#endif /* RMF_SEQUENCE_DECORATORS_H */
