/**
 *  \file RMF/shape_decorators.h
 *  \brief Helper functions for manipulating RMF files.
 *
 *  Copyright 2007-2013 IMP Inventors. All rights reserved.
 *
 */

#ifndef RMF_SHAPE_DECORATORS_H
#define RMF_SHAPE_DECORATORS_H

#include <RMF/config.h>
#include <RMF/infrastructure_macros.h>
#include <RMF/NodeHandle.h>
#include <RMF/FileHandle.h>
#include <RMF/Decorator.h>
#include <RMF/Factory.h>
#include <RMF/constants.h>
#include <RMF/internal/utility.h>
#include <RMF/internal/paths.h>
#include <boost/array.hpp>
#include <boost/lexical_cast.hpp>

RMF_ENABLE_WARNINGS
namespace RMF {

/** These particles have associated color information.

       See also Colored and ColoredConstFactory.
     */
    class ColoredConst:
        public Decorator<NodeConstHandle> {
    friend class ColoredConstFactory;
    std::string get_name() const {return "Colored";}
    private:
    typedef Decorator<NodeConstHandle> P;
    FloatKeys rgb_color_;
    ColoredConst(NodeConstHandle nh,
                  FloatKeys rgb_color):
       P(nh),
rgb_color_(rgb_color) {
    ;
    }
    public:
    /** The RGB color. Each component has a value in [0...1]. */
Floats get_rgb_color() const {
             return P::get_values(rgb_color_);
           }
    static std::string get_decorator_type_name() {
         return "ColoredConst";
    }
    };

    typedef std::vector<ColoredConst>
            ColoredConsts;

/** These particles have associated color information.

       See also ColoredConst and ColoredFactory.
     */
    class Colored:
        public Decorator<NodeHandle> {
    friend class ColoredFactory;
    std::string get_name() const {return "Colored";}
    private:
    typedef Decorator<NodeHandle> P;
    FloatKeys rgb_color_;
    Colored(NodeHandle nh,
                  FloatKeys rgb_color):
       P(nh),
rgb_color_(rgb_color) {
    ;
    }
    public:
    /** The RGB color. Each component has a value in [0...1]. */
Floats get_rgb_color() const {
             return P::get_values(rgb_color_);
           }
/** The RGB color. Each component has a value in [0...1]. */
    void set_rgb_color(const Floats &v) {
           P::set_values(rgb_color_, v);
        }
    static std::string get_decorator_type_name() {
         return "Colored";
    }
    };

    typedef std::vector<Colored>
            Coloreds;

/** Create decorators of type Colored.

       See also Colored and ColoredConstFactory.
    */
    class ColoredFactory:
       public Factory<FileHandle>
                 {
    private:
    typedef Factory<FileHandle> P;
    FloatKeys rgb_color_;
    public:
    ColoredFactory(FileHandle fh):
      P(){
    Category cat = fh.get_category("shape");
RMF_UNUSED(cat);
        Strings  rgb_color_names;
        rgb_color_names.push_back("rgb color red");
        rgb_color_names.push_back("rgb color green");
        rgb_color_names.push_back("rgb color blue");
      rgb_color_=fh.get_keys<FloatTraits>(cat,
        rgb_color_names);;
    }
    /** Get a Colored for nh.*/
    Colored get(NodeHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION||nh.get_type() == RMF::ORGANIZATIONAL||nh.get_type() == RMF::ALIAS||nh.get_type() == RMF::FEATURE||nh.get_type() == RMF::GEOMETRY), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Colored");;
      return Colored(nh, rgb_color_);
    }
    /** Check whether nh has all the attributes required to be a
        Colored.*/
    bool get_is(NodeHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION||nh.get_type() == RMF::ORGANIZATIONAL||nh.get_type() == RMF::ALIAS||nh.get_type() == RMF::FEATURE||nh.get_type() == RMF::GEOMETRY)
    && nh.get_has_value(rgb_color_[0]);
    }
    };

    typedef std::vector<ColoredFactory>
            ColoredFactories;

/** Create decorators of type Colored.

       See also ColoredConst and ColoredFactory.
    */
    class ColoredConstFactory:
       public Factory<FileConstHandle>
                 {
    private:
    typedef Factory<FileConstHandle> P;
    FloatKeys rgb_color_;
    public:
    ColoredConstFactory(FileConstHandle fh):
      P(){
    Category cat = fh.get_category("shape");
RMF_UNUSED(cat);
        Strings  rgb_color_names;
        rgb_color_names.push_back("rgb color red");
        rgb_color_names.push_back("rgb color green");
        rgb_color_names.push_back("rgb color blue");
      rgb_color_=fh.get_keys<FloatTraits>(cat,
        rgb_color_names);;
    }
    /** Get a ColoredConst for nh.*/
    ColoredConst get(NodeConstHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::REPRESENTATION||nh.get_type() == RMF::ORGANIZATIONAL||nh.get_type() == RMF::ALIAS||nh.get_type() == RMF::FEATURE||nh.get_type() == RMF::GEOMETRY), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Colored");;
      return ColoredConst(nh, rgb_color_);
    }
    /** Check whether nh has all the attributes required to be a
        ColoredConst.*/
    bool get_is(NodeConstHandle nh) const {
      return (nh.get_type() == RMF::REPRESENTATION||nh.get_type() == RMF::ORGANIZATIONAL||nh.get_type() == RMF::ALIAS||nh.get_type() == RMF::FEATURE||nh.get_type() == RMF::GEOMETRY)
    && nh.get_has_value(rgb_color_[0]);
    }
    };

    typedef std::vector<ColoredConstFactory>
            ColoredConstFactories;

/** A geometric ball.

       See also Ball and BallConstFactory.
     */
    class BallConst:
        public Decorator<NodeConstHandle> {
    friend class BallConstFactory;
    std::string get_name() const {return "Ball";}
    private:
    typedef Decorator<NodeConstHandle> P;
    FloatKeys coordinates_;
FloatKey radius_;
IndexKey type_;
    BallConst(NodeConstHandle nh,
                  FloatKeys coordinates,
FloatKey radius,
IndexKey type):
       P(nh),
coordinates_(coordinates),
radius_(radius),
type_(type) {
    ;
    }
    public:
    /** Coordinates of the center in angstroms. */
Floats get_coordinates() const {
         Floats ret(3);
           for (unsigned int i=0; i< 3; ++i) {
            ret[i]=get_node().get_value(coordinates_[i]);
           }
         return ret;
      }
/** The radius in angstroms. */
Float get_radius() const {
  try{
  return P::get_value(radius_);
  } RMF_DECORATOR_CATCH( );
}
/** The radius in angstroms. */
Floats get_all_radiuss() const {
  try {
  return P::get_all_values(radius_);
  } RMF_DECORATOR_CATCH( );
}
    static std::string get_decorator_type_name() {
         return "BallConst";
    }
    };

    typedef std::vector<BallConst>
            BallConsts;

/** A geometric ball.

       See also BallConst and BallFactory.
     */
    class Ball:
        public Decorator<NodeHandle> {
    friend class BallFactory;
    std::string get_name() const {return "Ball";}
    private:
    typedef Decorator<NodeHandle> P;
    FloatKeys coordinates_;
FloatKey radius_;
IndexKey type_;
    Ball(NodeHandle nh,
                  FloatKeys coordinates,
FloatKey radius,
IndexKey type):
       P(nh),
coordinates_(coordinates),
radius_(radius),
type_(type) {
    nh.set_value(type_, 0);;
    }
    public:
    /** Coordinates of the center in angstroms. */
Floats get_coordinates() const {
         Floats ret(3);
           for (unsigned int i=0; i< 3; ++i) {
            ret[i]=get_node().get_value(coordinates_[i]);
           }
         return ret;
      }
/** Coordinates of the center in angstroms. */
void set_coordinates(const Floats &v) {
             for (unsigned int i=0; i< 3; ++i) {
                get_node().set_value(coordinates_[i], v[i]);
             }
        }
/** The radius in angstroms. */
Float get_radius() const {
  try{
  return P::get_value(radius_);
  } RMF_DECORATOR_CATCH( );
}
/** The radius in angstroms. */
Floats get_all_radiuss() const {
  try {
  return P::get_all_values(radius_);
  } RMF_DECORATOR_CATCH( );
}
/** The radius in angstroms. */
void set_radius(Float v) {
  try {
  P::set_value(radius_,
            v);
  } RMF_DECORATOR_CATCH( );
}
    static std::string get_decorator_type_name() {
         return "Ball";
    }
    };

    typedef std::vector<Ball>
            Balls;

/** Create decorators of type Ball.

       See also Ball and BallConstFactory.
    */
    class BallFactory:
       public Factory<FileHandle>
                 {
    private:
    typedef Factory<FileHandle> P;
    FloatKeys coordinates_;
FloatKey radius_;
IndexKey type_;
    public:
    BallFactory(FileHandle fh):
      P(){
    Category cat = fh.get_category("shape");
RMF_UNUSED(cat);
        Strings  coordinates_names;
        coordinates_names.push_back("cartesian x");
        coordinates_names.push_back("cartesian y");
        coordinates_names.push_back("cartesian z");
      coordinates_=fh.get_keys<FloatTraits>(cat,
        coordinates_names);
radius_=fh.get_key<FloatTraits>(cat,
        "radius");
type_=fh.get_key<IndexTraits>(cat,
        "type");;
    }
    /** Get a Ball for nh.*/
    Ball get(NodeHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::GEOMETRY), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Ball");;
      return Ball(nh, coordinates_,
radius_,
type_);
    }
    /** Check whether nh has all the attributes required to be a
        Ball.*/
    bool get_is(NodeHandle nh) const {
      return (nh.get_type() == RMF::GEOMETRY)
    && nh.get_has_value(coordinates_[0])
    && nh.get_has_value(radius_)
    && nh.get_has_value(type_)
    && nh.get_value(type_)==0;
    }
    };

    typedef std::vector<BallFactory>
            BallFactories;

/** Create decorators of type Ball.

       See also BallConst and BallFactory.
    */
    class BallConstFactory:
       public Factory<FileConstHandle>
                 {
    private:
    typedef Factory<FileConstHandle> P;
    FloatKeys coordinates_;
FloatKey radius_;
IndexKey type_;
    public:
    BallConstFactory(FileConstHandle fh):
      P(){
    Category cat = fh.get_category("shape");
RMF_UNUSED(cat);
        Strings  coordinates_names;
        coordinates_names.push_back("cartesian x");
        coordinates_names.push_back("cartesian y");
        coordinates_names.push_back("cartesian z");
      coordinates_=fh.get_keys<FloatTraits>(cat,
        coordinates_names);
radius_=fh.get_key<FloatTraits>(cat,
        "radius");
type_=fh.get_key<IndexTraits>(cat,
        "type");;
    }
    /** Get a BallConst for nh.*/
    BallConst get(NodeConstHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::GEOMETRY), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Ball");;
      return BallConst(nh, coordinates_,
radius_,
type_);
    }
    /** Check whether nh has all the attributes required to be a
        BallConst.*/
    bool get_is(NodeConstHandle nh) const {
      return (nh.get_type() == RMF::GEOMETRY)
    && nh.get_has_value(coordinates_[0])
    && nh.get_has_value(radius_)
    && nh.get_has_value(type_)
    && nh.get_value(type_)==0;
    }
    };

    typedef std::vector<BallConstFactory>
            BallConstFactories;

/** A geometric cylinder.

       See also Cylinder and CylinderConstFactory.
     */
    class CylinderConst:
        public Decorator<NodeConstHandle> {
    friend class CylinderConstFactory;
    std::string get_name() const {return "Cylinder";}
    private:
    typedef Decorator<NodeConstHandle> P;
    FloatsKeys coordinates_;
FloatKey radius_;
IndexKey type_;
    CylinderConst(NodeConstHandle nh,
                  FloatsKeys coordinates,
FloatKey radius,
IndexKey type):
       P(nh),
coordinates_(coordinates),
radius_(radius),
type_(type) {
    ;
    }
    public:
    /** The coordinates of the endpoints in angstroms. The returned list has 3 components, a list of x coordinates, a list of y coordinates and a list of z coordinates. */
FloatsList get_coordinates() const {
         FloatsList ret(3);
           for (unsigned int i=0; i< 3; ++i) {
            ret[i]=get_node().get_value(coordinates_[i]);
           }
         return ret;
      }
/** The radius in angstroms. */
Float get_radius() const {
  try{
  return P::get_value(radius_);
  } RMF_DECORATOR_CATCH( );
}
/** The radius in angstroms. */
Floats get_all_radiuss() const {
  try {
  return P::get_all_values(radius_);
  } RMF_DECORATOR_CATCH( );
}
    static std::string get_decorator_type_name() {
         return "CylinderConst";
    }
    };

    typedef std::vector<CylinderConst>
            CylinderConsts;

/** A geometric cylinder.

       See also CylinderConst and CylinderFactory.
     */
    class Cylinder:
        public Decorator<NodeHandle> {
    friend class CylinderFactory;
    std::string get_name() const {return "Cylinder";}
    private:
    typedef Decorator<NodeHandle> P;
    FloatsKeys coordinates_;
FloatKey radius_;
IndexKey type_;
    Cylinder(NodeHandle nh,
                  FloatsKeys coordinates,
FloatKey radius,
IndexKey type):
       P(nh),
coordinates_(coordinates),
radius_(radius),
type_(type) {
    nh.set_value(type_, 1);;
    }
    public:
    /** The coordinates of the endpoints in angstroms. The returned list has 3 components, a list of x coordinates, a list of y coordinates and a list of z coordinates. */
FloatsList get_coordinates() const {
         FloatsList ret(3);
           for (unsigned int i=0; i< 3; ++i) {
            ret[i]=get_node().get_value(coordinates_[i]);
           }
         return ret;
      }
/** The coordinates of the endpoints in angstroms. The returned list has 3 components, a list of x coordinates, a list of y coordinates and a list of z coordinates. */
void set_coordinates(const FloatsList &v) {
             for (unsigned int i=0; i< 3; ++i) {
                get_node().set_value(coordinates_[i], v[i]);
             }
        }
/** The radius in angstroms. */
Float get_radius() const {
  try{
  return P::get_value(radius_);
  } RMF_DECORATOR_CATCH( );
}
/** The radius in angstroms. */
Floats get_all_radiuss() const {
  try {
  return P::get_all_values(radius_);
  } RMF_DECORATOR_CATCH( );
}
/** The radius in angstroms. */
void set_radius(Float v) {
  try {
  P::set_value(radius_,
            v);
  } RMF_DECORATOR_CATCH( );
}
    static std::string get_decorator_type_name() {
         return "Cylinder";
    }
    };

    typedef std::vector<Cylinder>
            Cylinders;

/** Create decorators of type Cylinder.

       See also Cylinder and CylinderConstFactory.
    */
    class CylinderFactory:
       public Factory<FileHandle>
                 {
    private:
    typedef Factory<FileHandle> P;
    FloatsKeys coordinates_;
FloatKey radius_;
IndexKey type_;
    public:
    CylinderFactory(FileHandle fh):
      P(){
    Category cat = fh.get_category("shape");
RMF_UNUSED(cat);
        Strings  coordinates_names;
        coordinates_names.push_back("cartesian xs");
        coordinates_names.push_back("cartesian ys");
        coordinates_names.push_back("cartesian zs");
      coordinates_=fh.get_keys<FloatsTraits>(cat,
        coordinates_names);
radius_=fh.get_key<FloatTraits>(cat,
        "radius");
type_=fh.get_key<IndexTraits>(cat,
        "type");;
    }
    /** Get a Cylinder for nh.*/
    Cylinder get(NodeHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::GEOMETRY), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Cylinder");;
      return Cylinder(nh, coordinates_,
radius_,
type_);
    }
    /** Check whether nh has all the attributes required to be a
        Cylinder.*/
    bool get_is(NodeHandle nh) const {
      return (nh.get_type() == RMF::GEOMETRY)
    && nh.get_has_value(coordinates_[0])
    && nh.get_has_value(radius_)
    && nh.get_has_value(type_)
    && nh.get_value(type_)==1;
    }
    };

    typedef std::vector<CylinderFactory>
            CylinderFactories;

/** Create decorators of type Cylinder.

       See also CylinderConst and CylinderFactory.
    */
    class CylinderConstFactory:
       public Factory<FileConstHandle>
                 {
    private:
    typedef Factory<FileConstHandle> P;
    FloatsKeys coordinates_;
FloatKey radius_;
IndexKey type_;
    public:
    CylinderConstFactory(FileConstHandle fh):
      P(){
    Category cat = fh.get_category("shape");
RMF_UNUSED(cat);
        Strings  coordinates_names;
        coordinates_names.push_back("cartesian xs");
        coordinates_names.push_back("cartesian ys");
        coordinates_names.push_back("cartesian zs");
      coordinates_=fh.get_keys<FloatsTraits>(cat,
        coordinates_names);
radius_=fh.get_key<FloatTraits>(cat,
        "radius");
type_=fh.get_key<IndexTraits>(cat,
        "type");;
    }
    /** Get a CylinderConst for nh.*/
    CylinderConst get(NodeConstHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::GEOMETRY), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Cylinder");;
      return CylinderConst(nh, coordinates_,
radius_,
type_);
    }
    /** Check whether nh has all the attributes required to be a
        CylinderConst.*/
    bool get_is(NodeConstHandle nh) const {
      return (nh.get_type() == RMF::GEOMETRY)
    && nh.get_has_value(coordinates_[0])
    && nh.get_has_value(radius_)
    && nh.get_has_value(type_)
    && nh.get_value(type_)==1;
    }
    };

    typedef std::vector<CylinderConstFactory>
            CylinderConstFactories;

/** A geometric line setgment.

       See also Segment and SegmentConstFactory.
     */
    class SegmentConst:
        public Decorator<NodeConstHandle> {
    friend class SegmentConstFactory;
    std::string get_name() const {return "Segment";}
    private:
    typedef Decorator<NodeConstHandle> P;
    FloatsKeys coordinates_;
IndexKey type_;
    SegmentConst(NodeConstHandle nh,
                  FloatsKeys coordinates,
IndexKey type):
       P(nh),
coordinates_(coordinates),
type_(type) {
    ;
    }
    public:
    /** The coordinates of the endpoints in angstroms. The returned list has 3 components, a list of x coordinates, a list of y coordinates and a list of z coordinates. */
FloatsList get_coordinates() const {
         FloatsList ret(3);
           for (unsigned int i=0; i< 3; ++i) {
            ret[i]=get_node().get_value(coordinates_[i]);
           }
         return ret;
      }
    static std::string get_decorator_type_name() {
         return "SegmentConst";
    }
    };

    typedef std::vector<SegmentConst>
            SegmentConsts;

/** A geometric line setgment.

       See also SegmentConst and SegmentFactory.
     */
    class Segment:
        public Decorator<NodeHandle> {
    friend class SegmentFactory;
    std::string get_name() const {return "Segment";}
    private:
    typedef Decorator<NodeHandle> P;
    FloatsKeys coordinates_;
IndexKey type_;
    Segment(NodeHandle nh,
                  FloatsKeys coordinates,
IndexKey type):
       P(nh),
coordinates_(coordinates),
type_(type) {
    nh.set_value(type_, 1);;
    }
    public:
    /** The coordinates of the endpoints in angstroms. The returned list has 3 components, a list of x coordinates, a list of y coordinates and a list of z coordinates. */
FloatsList get_coordinates() const {
         FloatsList ret(3);
           for (unsigned int i=0; i< 3; ++i) {
            ret[i]=get_node().get_value(coordinates_[i]);
           }
         return ret;
      }
/** The coordinates of the endpoints in angstroms. The returned list has 3 components, a list of x coordinates, a list of y coordinates and a list of z coordinates. */
void set_coordinates(const FloatsList &v) {
             for (unsigned int i=0; i< 3; ++i) {
                get_node().set_value(coordinates_[i], v[i]);
             }
        }
    static std::string get_decorator_type_name() {
         return "Segment";
    }
    };

    typedef std::vector<Segment>
            Segments;

/** Create decorators of type Segment.

       See also Segment and SegmentConstFactory.
    */
    class SegmentFactory:
       public Factory<FileHandle>
                 {
    private:
    typedef Factory<FileHandle> P;
    FloatsKeys coordinates_;
IndexKey type_;
    public:
    SegmentFactory(FileHandle fh):
      P(){
    Category cat = fh.get_category("shape");
RMF_UNUSED(cat);
        Strings  coordinates_names;
        coordinates_names.push_back("cartesian xs");
        coordinates_names.push_back("cartesian ys");
        coordinates_names.push_back("cartesian zs");
      coordinates_=fh.get_keys<FloatsTraits>(cat,
        coordinates_names);
type_=fh.get_key<IndexTraits>(cat,
        "type");;
    }
    /** Get a Segment for nh.*/
    Segment get(NodeHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::GEOMETRY), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Segment");;
      return Segment(nh, coordinates_,
type_);
    }
    /** Check whether nh has all the attributes required to be a
        Segment.*/
    bool get_is(NodeHandle nh) const {
      return (nh.get_type() == RMF::GEOMETRY)
    && nh.get_has_value(coordinates_[0])
    && nh.get_has_value(type_)
    && nh.get_value(type_)==1;
    }
    };

    typedef std::vector<SegmentFactory>
            SegmentFactories;

/** Create decorators of type Segment.

       See also SegmentConst and SegmentFactory.
    */
    class SegmentConstFactory:
       public Factory<FileConstHandle>
                 {
    private:
    typedef Factory<FileConstHandle> P;
    FloatsKeys coordinates_;
IndexKey type_;
    public:
    SegmentConstFactory(FileConstHandle fh):
      P(){
    Category cat = fh.get_category("shape");
RMF_UNUSED(cat);
        Strings  coordinates_names;
        coordinates_names.push_back("cartesian xs");
        coordinates_names.push_back("cartesian ys");
        coordinates_names.push_back("cartesian zs");
      coordinates_=fh.get_keys<FloatsTraits>(cat,
        coordinates_names);
type_=fh.get_key<IndexTraits>(cat,
        "type");;
    }
    /** Get a SegmentConst for nh.*/
    SegmentConst get(NodeConstHandle nh) const {
      RMF_USAGE_CHECK((nh.get_type() == RMF::GEOMETRY), std::string("Bad node type. Got \"")
                                      + boost::lexical_cast<std::string>(nh.get_type())
                                      + "\" in decorator type  Segment");;
      return SegmentConst(nh, coordinates_,
type_);
    }
    /** Check whether nh has all the attributes required to be a
        SegmentConst.*/
    bool get_is(NodeConstHandle nh) const {
      return (nh.get_type() == RMF::GEOMETRY)
    && nh.get_has_value(coordinates_[0])
    && nh.get_has_value(type_)
    && nh.get_value(type_)==1;
    }
    };

    typedef std::vector<SegmentConstFactory>
            SegmentConstFactories;

} /* namespace RMF */
RMF_DISABLE_WARNINGS

#endif /* RMF_SHAPE_DECORATORS_H */
